@@ -9,11 +9,9 @@ mod tests;
 
 use self::runtime::CallFrame;
 use crate::{
-    binary_compare,
     collections::key_map::{Handle, KeyMap},
     instruction::instruction_span,
     instruction::Instruction,
-    pop_stack,
     prelude::*,
     value::Value,
     VariableId,
@@ -63,6 +61,7 @@ impl<'a, Aux> Vm<'a, Aux> {
         self.runtime_data.global_vars.get(name.0 as usize).cloned()
     }
 
+    #[must_use]
     pub fn with_max_iter(mut self, max_iter: u64) -> Self {
         self.max_instr = max_iter;
         self
@@ -146,8 +145,27 @@ impl<'a, Aux> Vm<'a, Aux> {
         self.runtime_data.init_table()
     }
 
+    /// Initializes a new string owned by this VM instance
+    pub fn init_string(&mut self, payload: &str) -> Result<StrPointer, ExecutionError> {
+        unsafe {
+            let layout = std::alloc::Layout::from_size_align(4 + payload.len(), 4).unwrap();
+            let mut ptr = self
+                .runtime_data
+                .memory
+                .alloc(layout)
+                .map_err(|_| ExecutionError::OutOfMemory)?;
+
+            let result: *mut u8 = ptr.as_mut();
+            std::ptr::write(result as *mut u32, payload.len() as u32);
+            std::ptr::copy(payload.as_ptr(), result.add(4), payload.len());
+
+            Ok(StrPointer(ptr.as_ptr()))
+        }
+    }
+
     /// This mostly assumes that program is valid, produced by the compiler.
     /// As such running non-compiler emitted programs is very un-safe
+    #[inline(never)]
     pub fn run(&mut self, program: &CaoProgram) -> Result<(), ExecutionError> {
         self.runtime_data
             .call_stack
@@ -160,12 +178,13 @@ impl<'a, Aux> Vm<'a, Aux> {
         let mut remaining_iters = self.max_instr;
 
         let mut instr_ptr = 0;
+        let bytecode_ptr = program.bytecode.as_ptr();
         while instr_ptr < len {
             remaining_iters -= 1;
             if remaining_iters == 0 {
                 return Err(ExecutionError::Timeout);
             }
-            let instr: u8 = unsafe { *program.bytecode.as_ptr().add(instr_ptr) };
+            let instr: u8 = unsafe { *bytecode_ptr.add(instr_ptr) };
             let instr: Instruction = unsafe { transmute(instr) };
             instr_ptr += 1;
             debug!(
@@ -251,8 +270,8 @@ impl<'a, Aux> Vm<'a, Aux> {
                     instr_ptr = pos as usize;
                 }
                 Instruction::SwapLast => {
-                    let b = pop_stack!(self);
-                    let a = pop_stack!(self);
+                    let b = self.stack_pop();
+                    let a = self.stack_pop();
                     // we popped two values, we know that the stack has capacity for 2 ..
                     self.stack_push(b).unwrap();
                     self.stack_push(a).unwrap();
@@ -288,9 +307,7 @@ impl<'a, Aux> Vm<'a, Aux> {
                     )?;
                 }
                 Instruction::Pop => {
-                    if !self.runtime_data.stack.is_empty() {
-                        self.runtime_data.stack.pop();
-                    }
+                    self.stack_pop();
                 }
                 Instruction::CallLane => {
                     instr_execution::instr_jump(&mut instr_ptr, program, &mut self.runtime_data)?;
@@ -337,10 +354,10 @@ impl<'a, Aux> Vm<'a, Aux> {
                 Instruction::Sub => self.binary_op(|a, b| a - b)?,
                 Instruction::Mul => self.binary_op(|a, b| a * b)?,
                 Instruction::Div => self.binary_op(|a, b| a / b)?,
-                Instruction::Equals => binary_compare!(self, ==, false),
-                Instruction::NotEquals => binary_compare!(self, !=, true),
-                Instruction::Less => binary_compare!(self, <, false),
-                Instruction::LessOrEq => binary_compare!(self, <=, false),
+                Instruction::Equals => self.binary_op(|a, b| (a == b).into())?,
+                Instruction::NotEquals => self.binary_op(|a, b| (a != b).into())?,
+                Instruction::Less => self.binary_op(|a, b| (a < b).into())?,
+                Instruction::LessOrEq => self.binary_op(|a, b| (a <= b).into())?,
                 Instruction::StringLiteral => {
                     instr_execution::instr_string_literal(self, &mut instr_ptr, program)?
                 }
@@ -354,12 +371,10 @@ impl<'a, Aux> Vm<'a, Aux> {
         Err(ExecutionError::UnexpectedEndOfInput)
     }
 
-    fn binary_op<F>(&mut self, op: F) -> Result<(), ExecutionError>
-    where
-        F: Fn(Value, Value) -> Value,
-    {
-        let b = pop_stack!(self);
-        let a = pop_stack!(self);
+    #[inline]
+    fn binary_op(&mut self, op: fn(Value, Value) -> Value) -> Result<(), ExecutionError> {
+        let b = self.stack_pop();
+        let a = self.stack_pop();
 
         self.runtime_data
             .stack
